/***********************************************************************/
/*                                                                     */
/*                               Pio (C)                               */
/*                                                                     */
/*                    Polyglot Input/Output Library                    */
/*                                                                     */
/*               EPI Pomdapi, INRIA Paris - Rocquencourt               */
/*                                                                     */
/*  Copyright 2011-2014 INRIA.                                         */
/*  Institut National de Recherche en Informatique et en Automatique.  */
/*  All rights reserved. This file is distributed only by permission.  */
/*                                                                     */
/*  Bruno Verlyck <Bruno.Verlyck@inria.fr>                             */
/*                                                                     */
/***********************************************************************/

#include "common.h"
#include "debug_io.h"
#include "pio_ibuff.h"
#include "typed_value.h"
#include "read_typed_value.h"
#include "communication.h"
#include "read_communication.h"

#define xdebug(fun_name, msg) pio_debug (__FILE__ "::" fun_name, msg)
#define DEB_BEGIN xdebug (funname, "start reading")
#define DEB_END xdebug (funname, "finish reading")

#define funname "r_values"
static TV_t **r_values (pio_input_buffer *ib, size_t n) {
  DEB_BEGIN;
  if (n == 0) {
    xdebug (funname, "0 value(s) read");
    return NULL;
  }

  TV_t **values = calloc (n + 1, sizeof (*values));
  if (values == NULL)
      failwith (__FILE__ "::" funname ": out of memory");
  for (size_t i = 0; i < n; i++) {
    values[i] = read_typed_value (ib);
  }
  values[n] = NULL;			/* eoa(rray) marker (needed by receive_values) */
  /* let's cheat a little with failbuf */
  xdebug (funname, (snprintf (failbuf, sizeof (failbuf), "%li value(s) read", (long int)n), failbuf));
  return values;
}
#undef funname

static const char *read_name (pio_input_buffer *ib, char *name_name) {
  struct string name = read_delimited_string (ib, 2);
  if (name.ssize != strlen (name.str)) {
    raise_wrong_communication ((snprintf (failbuf, sizeof (failbuf), 
					  "%s name starting with `%s' contains NUL(s)", name_name, name.str),
				failbuf));
  }
  ibuf_scanf0 (ib, "\n");
  return name.str;
}

#define funname "read_communication"
COM_t *unsafe_read_communication (pio_input_buffer *ib) {
  DEB_BEGIN;
  ibuf_scanf0 (ib, "(\n");
  xdebug (funname, "\"(\" marker read");
  char t; size_t n;
  ibuf_scanf (ib, 2, "%%%c <%ld> ", &t, &n);
  COM_t *res;
  switch (t) {
  case 'p':
    xdebug (funname, "phrase tag read");
    res = make_communication (COM_Phrase);
    res->vs.phrase_values = r_values (ib, n);
    break;
  case 't':
    xdebug (funname, "task tag read");
    res = make_communication (COM_Task);
    res->vs.task.t_name = read_name (ib, "task");
    res->vs.task.t_values = r_values (ib, n);
    break;
  case 'r':
    xdebug (funname, "result tag read");
    res = make_communication (COM_Result);
    res->vs.result_values = r_values (ib, n);
    break;
  case 'e':
    xdebug (funname, "error tag read");
    res = make_communication (COM_Error);
    res->vs.error_values = r_values (ib, n);
    break;
  case 's':
    xdebug (funname, "service tag read");
    res = make_communication (COM_Service);
    res->vs.svc.srv_name = service_name_of_string (read_name (ib, "service"));
    res->vs.svc.srv_values = r_values (ib, n);
    break;
  default:
    raise_wrong_communication ((snprintf (failbuf, sizeof (failbuf),
					  "Unknown communication kind ``%%%c''", t),
				failbuf));
  }
  res->com_nvalues = n;
  char d;
  ibuf_scanf (ib, 1, "%c", &d);		/* %0c fait %1c contrairement  OCaml */
  if (d != ')') {
    raise_wrong_communication ((snprintf (failbuf, sizeof (failbuf),
					  "Unknown end of communication character ``%c''", d),
				failbuf));
  }
  ibuf_scanf0 (ib, "\n\n");
  xdebug (funname, "\")\" marker read");
  DEB_END;
  return res;
}

/* Any modification here must be reported in io.c::gen_receive */
/* XXXX should save associated msg */
COM_t *read_communication (pio_input_buffer *ib) {
  jmp_buf rc_buf;
  jmp_buf *sv_wrong_communication = wrong_communication_label;
  wrong_communication_label = &rc_buf;
  jmp_buf *sv_eof_bscanf_label = eof_bscanf_label;
  eof_bscanf_label = &rc_buf;
  jmp_buf *sv_fail_label = fail_label;
  fail_label = &rc_buf;
  COM_t *res;
  int come_from = (setjmp (rc_buf));
  if (!come_from) {
    res = unsafe_read_communication (ib);
  } else {
    const char *exc_msg =
      (come_from == (int)((long)(&wrong_communication_label)))
      ? wrong_communication_msg
      : (come_from == (int)((long)(&eof_bscanf_label)))
      ? "Eof in ibuf_scanf"
      : (come_from == (int)((long)(&fail_label)))
      ? failure_msg
      : (assert (0), "unknown exception in read_communication");
    res = make_communication (COM_wrong_communication);
    res->vs.svc.srv_name = service_name_of_string (exc_msg);
    res->com_nvalues = 1;
    TV_t *exc_typed_value = TV_of_chars (exc_msg);
    TV_t **values = calloc (1, sizeof exc_typed_value);
    values[0] = exc_typed_value;
    res->vs.svc.srv_values = values;
  }
  wrong_communication_label = sv_wrong_communication;
  eof_bscanf_label = sv_eof_bscanf_label;
  fail_label = sv_fail_label;
  return res;
}
