(***********************************************************************)
(*                                                                     *)
(*                             Pio (OCaml)                             *)
(*                                                                     *)
(*                    Polyglot Input/Output Library                    *)
(*                                                                     *)
(*               EPI Pomdapi, INRIA Paris - Rocquencourt               *)
(*                                                                     *)
(*  Copyright 2005-2014 INRIA.                                         *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved. This file is distributed only by permission.  *)
(*                                                                     *)
(*  Francois Clement <Francois.Clement@inria.fr>                       *)
(*  Pierre Weis <Pierre.Weis@inria.fr>                                 *)
(*                                                                     *)
(***********************************************************************)

(* $Id: read_communication.ml,v 1.12 2014/01/10 19:22:10 fclement Exp $ *)

open Typed_value;;
open Communication;;

let debug fun_name msg =
  Options_pio.debug (Printf.sprintf "Read_communication.%s" fun_name) msg
;;

let fatal_error fun_name msg =
  Options_pio.fatal_error
    (Printf.sprintf "Read_communication.%s" fun_name) msg 2
;;

let r_values ib n =
  assert (debug "r_values" "start reading");
  if n = 0 then
    begin
      assert (debug "r_values" "0 value(s) read");
      [| |]
    end
  else
    begin
      let vals = Array.make n (Scalar (Sint 0)) in
      for i = 0 to n - 1 do
        vals.(i) <- Read_typed_value.read_typed_value ib
      done;
      assert (debug "r_values" (Printf.sprintf "%i value(s) read" n));
      vals
    end
;;

let unsafe_read_communication ib =
  assert (debug "read_communication" "start reading");
  Scanf.bscanf ib "(\n" ();
  assert (debug "read_communication" "\"(\" marker read");
  Scanf.bscanf ib "%%%c <%d> " (fun t n ->
    let com =
      match t with
      | 'p' ->
          assert (debug "read_communication" "phrase tag read");
          Phrase (r_values ib n)
      | 't' ->
          assert (debug "read_communication" "task tag read");
          Scanf.bscanf ib "%S\n" (fun task_name ->
            assert (debug "read_communication" "task name read");
            let vals = r_values ib n in
            Task (task_name, vals))
      | 'r' ->
          assert (debug "read_communication" "result tag read");
          Result (r_values ib n)
      | 'e' ->
          assert (debug "read_communication" "error tag read");
          Error (r_values ib n)
      | 's' ->
          assert (debug "read_communication" "service tag read");
          Scanf.bscanf ib "%S\n" (fun sname ->
            assert (debug "read_communication" "service name read");
            let service_name = service_name_of_string sname in
            let vals = r_values ib n in
            Service (service_name, vals))
      | c ->
        raise
          (Typed_value.Wrong_communication
             (Printf.sprintf "Unknown communication kind ``%%%c''" c)) in
    Scanf.bscanf ib "%0c" (function
      | (* ( *) ')' ->
        begin
          Scanf.bscanf ib ")\n\n" ();
          assert (debug "read_communication" "\")\" marker read");
          assert (debug "read_communication" "finish reading");
          com
        end
      | c ->
        raise
          (Typed_value.Wrong_communication
            (Printf.sprintf
               "Unknown end of communication character ``%c''" c))))
;;

let read_communication ib =
  try unsafe_read_communication ib with
  (* Exceptions from module [Scanf] are caught here to return a service
     communication reporting the error. *)
  | Typed_value.Wrong_communication _
  | Scanf.Scan_failure _
  | Failure _
  | End_of_file as x ->
    Service (WrongCommunication, [| Scalar (Sstring (Printexc.to_string x)) |])
  (* Invalid_argument s reveals an error in the implementation of the library
     and should not be caught. *)
  | x -> fatal_error "read_communication" (Printexc.to_string x)
;;
