(***********************************************************************)
(*                                                                     *)
(*                                Sklml                                *)
(*                                                                     *)
(*                 The OCaml parallel skeleton system                  *)
(*                                                                     *)
(*               EPI Pomdapi, INRIA Paris - Rocquencourt               *)
(*                                                                     *)
(*  Copyright 2007-2014 INRIA.                                         *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved. This file is distributed only by permission.  *)
(*                                                                     *)
(*  Quentin Carbonneaux <Quentin.Carbonneaux@inria.fr>                 *)
(*  Francois Clement <Francois.Clement@inria.fr>                       *)
(*  Pierre Weis <Pierre.Weis@inria.fr>                                 *)
(*                                                                     *)
(***********************************************************************)

(* $Id: main.ml,v 1.11 2014/01/21 09:21:18 fclement Exp $ *)

open Skl;;
open Options;;

let stream_of_matrix mat =
  let (count, mx) = (ref 0, Array.length mat) in
  let gen () =
    if !count >= mx then None else
    let line = mat.(!count) in
    incr count; Some line in
  Sklstream.of_fun gen
;;

let vector_of_stream s = Array.of_list (Sklstream.to_list s);;

let predicate niter =
  let count = ref niter in
  fun _ -> decr count; !count <> 0
;;

(* The atomic skeletons. *)

let scalar_product_skl = skl v0 -> Helper.scalar_prod v0;;

let mk_pvect_skl = skl () -> Tridiag.mk_pvect;;

let pmult_skl = skl tmat -> Tridiag.pmult tmat;;

let predicate_skl = skl niter -> predicate niter;;

let display_skl =
  skl () ->
    fun x ->
      Helper.IO.output_raw stdout x;
      Printf.printf "\n%!";
      x
;;

(* The full computational skeleton expressions. *)

let matvec_skl vec nw = farm (scalar_product_skl vec, nw);;

let tridiag_skl nw niter tmat =
  loop
    (predicate_skl niter, (* do { } while (--count <> 0); *)
     mk_pvect_skl () |||
     farm_vector (pmult_skl tmat, nw) |||
     display_skl ())
;;

let main { Parfun.parfun = parfun; } =
  Ucommand_option.parse ();
  let { Options.number_of_workers = nw;
        vector_file = vf;
        matrix_file = mf;
        running_mode = mode; } = Options.get_options () in
  match mode with
  (* First demo, simple parallel multiplication *)
  | Std ->
      let (mat, vec) = Helper.fetch_std_data mf vf in
      let input_stream = stream_of_matrix mat in
      let output_stream = parfun (matvec_skl vec nw) input_stream in
      Helper.IO.output_vector stdout (vector_of_stream output_stream);
  (* Second demo, triangular matrix multiplication iterated *)
  | Tri ->
      let (mat, vec) = Helper.fetch_tri_data mf vf in
      let input_stream = Sklstream.singleton (Tridiag.split_vect 5 vec) in
      let output_stream = parfun (tridiag_skl nw 15 mat) input_stream in
      Sklstream.iter ignore  output_stream;
;;

Options.add_options ();;
Parfun.pardo main;;
