(***********************************************************************)
(*                                                                     *)
(*                                Sklml                                *)
(*                                                                     *)
(*                 The OCaml parallel skeleton system                  *)
(*                                                                     *)
(*               EPI Pomdapi, INRIA Paris - Rocquencourt               *)
(*                                                                     *)
(*  Copyright 2007-2014 INRIA.                                         *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved. This file is distributed only by permission.  *)
(*                                                                     *)
(*  Quentin Carbonneaux <Quentin.Carbonneaux@inria.fr>                 *)
(*  Francois Clement <Francois.Clement@inria.fr>                       *)
(*  Pierre Weis <Pierre.Weis@inria.fr>                                 *)
(*                                                                     *)
(***********************************************************************)

(* $Id: sklstream.mli,v 1.35 2014/01/21 09:21:15 fclement Exp $ *)

(** Streams are used to modelize the behaviour of computational nodes.

 Computational nodes are a kind of light servers or processes or threads, that
 get there input from the wires on the network and send their results on the
 wires.

 Since there is no thunk that can be passed on the wires, there are no lazy
 values in the streams. Networks feature delay in values availability, hence
 there is a kind of lazy evaluation in networks that we modelize via lazy
 evaluation in the stream data structure.

 In networks, even to get the first element of a stream, we have to explicitely
 ask and wait; hence a stream is a lazy data structure that has to be forced to
 get its head or even to know that it is indeed empty.

 We have:
   [ map_vector :
      ('a -> 'b) skel -> ('a vect, 'b vect) skel ].
 Or equivalently:
   [ map_vector :
      ('a stream -> 'b stream) -> ('a vect) stream -> ('b vect) stream ].
 Given the following semantics:
   [ map_vector skl \{| v1; ... vn; |\} =
       \{| to_vector (skl (to_stream v1)); ...;
           to_vector (skl (to_stream vn)); |\}. ]
 To implement [map_vector], we build a farm that computes the resulting
 vector for each vector of the input stream, using the [skl] argument as
 worker.
 As a consequence, the workers of the same farm are used for each element
 of each vector;
 and since each vector is sent to the wires as a (sub) stream of vector
 elements, each farm worker must apply its skeleton to a sub stream and
 send the resulting sub stream to be collected and converted back into a
 vector that is the next element of the stream result.

 Hence we need:
 [vector_to_substream : 'a array -> 'a sub_stream]
 [substream_to_vector : 'a sub_stream -> 'a array]

 And [skl : 'a sub_stream -> 'b sub_stream].

 A computational node ends when it encounters the end of stream marker
 [Nil] (which is the end of lazy list).
 It also must propagate the end of stream marker along the wires before dying.
*)

(** {6 Stream definition} *)

type 'a stream;;
(** The type of streams that run on the wires in parallel or
  are infinite data structures of lazy streams with ['a packet] elements.
*)

type 'a t = 'a stream;;
(** An alias for the [stream] type constructor. *)

(** {6 Stream construction functions} *)

val singleton : 'a -> 'a stream;;
(** [singleton x] returns the stream that has only [x] as element. *)

val head : 'a stream -> 'a;;
(** [head s] returns the first element of stream [s].
  Raises [End_of_stream], if [s] is empty. *)

(** {6 Stream generation/consumption utilities} *)

val of_fun : (unit -> 'a option) -> 'a stream;;
(** Generate a lazy stream from a function.
  The stream is closed with the value [None]. *)

val to_fun : 'a stream -> (unit -> 'a option);;
(** Generate the function that successively gives the elements of a lazy
  stream. *)

val of_list : 'a list -> 'a stream;;
(** Generate a (finite) lazy stream from a list. *)

val to_list : 'a stream -> 'a list;;
(** Generate the list of the elements of a (finite) lazy stream.
   This operation completely unfreezes the stream argument. *)

val of_vect : 'a array -> 'a stream;;
(** Generate a (finite) lazy stream from a vector. *)

val to_vect : 'a stream -> 'a array;;
(** Generate the vector of the elements of a (finite) lazy stream.
   This operation completely unfreezes the stream argument. *)

val of_channel : (in_channel -> 'a) -> in_channel -> 'a stream;;
(** Generate a lazy stream from an input channel. *)

val of_channel_by_lines : in_channel -> string stream;;
(** Generate a lazy stream from the lines of an input channel. *)

val of_channel_by_chars : in_channel -> char stream;;
(** Generate a lazy stream from the chars of an input channel. *)

(** {6 Operations on streams} *)

val join : ('a stream * 'a stream) -> 'a stream;;
(** Merges two lazy streams into a single stream with intermixed elements from
    both streams in a fair way (fairly peeks up alternatively elements from
    both streams). *)

val fork : 'a stream -> ('a stream * 'a stream);;
(** Splits a stream into two. *)

val in_par :
  (('a stream -> 'b stream) * ('a stream -> 'b stream)) ->
  'a stream -> 'b stream
;;

val split : ('a * 'b) stream -> 'a stream * 'b stream;;

val combine : 'a stream -> 'b stream -> ('a * 'b) stream;;

val in_prod :
  ('a stream -> 'b stream) * ('c stream -> 'd stream) ->
  ('a * 'c) stream -> ('b * 'd) stream
;;

type ('a, 'b) sum = Inl of 'a | Inr of 'b;;
(** This is the type of data handled by sum functions. *)

type control = private C_left | C_right;;
(** This is a private type storing information used internally by
    sum handling functions. *)

val sum_split :
  (('a, 'b) sum) stream -> (('a stream * 'b stream) * control stream)
;;
(** This splits a ['a + 'b] stream into two streams containing
    respectively elements of type ['a] and elements of type ['b]. A
    control stream is generated, it is used by [sum_combine]. *)

val sum_combine :
  (('a stream * 'a stream) * control stream) -> 'a stream
;;
(** This merges two streams of the same type using the control
    information stored in the control stream. *)

val in_sum :
  (('a stream -> 'c stream) * ('b stream -> 'c stream)) ->
  (('a, 'b) sum) stream -> 'c stream
;;

val fork_rails : int -> 'a stream -> 'a stream array;;
(** Splits a stream into n streams. *)

val join_rails : 'a stream array -> 'a stream;;
(** Merges len streams into one stream with intermixed elements. *)

val on_rails : ('a stream -> 'b stream) array -> 'a stream -> 'b stream;;

val lift : ('a -> 'b) -> 'a stream -> 'b stream;;
(** [lift f] lift the function [f] to a function on streams. *)
val unlift : ('a stream -> 'b stream) -> ('a -> 'b);;
(** [unlift fs] does the converse operation of [lift]: the function [fs]
  operating on streams is turned into a regular function. *)

val map : ('a -> 'b) -> 'a stream -> 'b stream;;
(** [map f s] maps function [f] on stream [s]:
    if [s] = [x1; x2; ...; xn] then [map f s] is the stream
    [f x1; f x2; ...; f xn]. *)

val foldl : ('a -> 'b -> 'a) -> 'a -> 'b stream -> 'a stream;;
(** Given a binary, associative, and commutative operator [op],
    its neutral element [zero] and a stream [s] = [x1; x2; ...; xn] then
    [fold_left op zero s] evaluates to
    [op (x1, op (x2, ..., (op (xn, zero))))]. *)

val loop : ('a stream -> bool stream) -> ('a stream -> 'a stream) ->
  'a stream -> 'a stream;;
(** This function implements the [loop] paradigm at the stream level:
   given two functions on streams (a predicate [p] and an effective
   function [f]) it returns the stream of fixpoints of [f] according
   to [p] using the elements of the input streams as
   initialization. *)

val iter : ('a -> unit) -> 'a stream -> unit;;
(** Applies a function to each element of a (finite) lazy stream.
   This operation completely unfreezes the stream argument. *)

(** {6 Serialization/deserialization for lists and vectors} *)

val serialize_stream_vector : 'a array stream -> 'a stream;;
val unserialize_stream_vector : 'a stream -> 'a array stream;;
(** The serialization/unserialization operations for vectors generalized to
  lazy streams of vectors. *)

val of_sarray : 'a array stream -> 'a stream;;
val to_sarray : 'a stream -> 'a array stream;;
(** Synonyms of the two previous functions. *)

(**/**)

(** {6 IO function} *)

val of_to_io_channel : (IO.ic * IO.oc) -> 'a stream -> 'b stream;;
val to_io_channel : 'a stream -> IO.oc -> unit;;
val of_io_channel : IO.ic -> 'a stream;;
val stream_to_io_channel : 'a stream -> IO.oc -> 'a stream;;

(** {6 DEBUGGING} *)

val print_stream :
  (Format.formatter -> 'a -> unit) ->
  Format.formatter -> 'a stream -> unit
;;

val eq_stream : 'a stream -> 'a stream -> bool
;;
