(***********************************************************************)
(*                                                                     *)
(*                                Sklml                                *)
(*                                                                     *)
(*                 The OCaml parallel skeleton system                  *)
(*                                                                     *)
(*               EPI Pomdapi, INRIA Paris - Rocquencourt               *)
(*                                                                     *)
(*  Copyright 2007-2014 INRIA.                                         *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved. This file is distributed only by permission.  *)
(*                                                                     *)
(*  Quentin Carbonneaux <Quentin.Carbonneaux@inria.fr>                 *)
(*  Francois Clement <Francois.Clement@inria.fr>                       *)
(*  Pierre Weis <Pierre.Weis@inria.fr>                                 *)
(*                                                                     *)
(***********************************************************************)

(* $Id: main.ml,v 1.18 2014/01/21 09:21:16 fclement Exp $ *)

(* Sklml parallel launcher daemon.
  This independant program can run in two modes:
  master or slave.  If it is in master mode it will be contacted by a
  master sklml program to create a computational network. Its
  different slaves will let him spawn the nodes of the computational
  network on several processors. If it runs in slave mode, it will be
  connected to a master daemon and spawn processes on its processor. *)

(* If the option [-master N] where N is an integer is given, it starts
  as a master node waiting for [N] slaves. If it is started with the
  option [-slave ADDR] where [ADDR] is a valid IPv4 address it will
  connect to the master node listenning on [ADDR] and act for it. *)

type mode =
   | Master of Master.node
   | Slave of Slave.node
;;

let get_usage () =
  "sklmlpard [OPTIONS] MODE\n\
   \n\
   Options for master mode only are tagged (M).\n\
   Options for slave mode only are tagged (S).\n\
   One and only one of the arguments tagged (MODE) must appear at the end of\n\
   the command line.\n"
;;

let make_default_master () =
  let local = Unix.inet_addr_loopback in
  {
    Master.local_addr = local;
    Master.nslaves = 1;
    Master.skl_port = 0;
    Master.slave_port = 0;
  }
;;

let make_default_slave () =
  let local = Unix.inet_addr_loopback in
  let addr = Unix.ADDR_INET (local, 0) in
  {
    Slave.color = 1;
    Slave.link = true;
    Slave.exe_path = "./skl.bin";
    Slave.local_addr = local;
    Slave.master = addr; (* Dummy address here *)
  }
;;

let parse_args argv =
  let mode = ref None in
  let master_mode = ref (make_default_master ())
  and slave_mode = ref (make_default_slave ()) in
  let (mkmaster, mkslave) =
    (fun x ->
      if x < 1 then raise (Arg.Bad "") else
      mode := Some (Master
        {!master_mode with Master.nslaves = x})
    ),
    (fun x ->
      mode := Some (Slave
        {!slave_mode with Slave.master = IO.read_sockaddr x})
    ) in
  let (set_slave_port, set_skl_port) =
    (fun port ->
      if port < 0 then raise (Arg.Bad "") else
      master_mode := {!master_mode with Master.slave_port = port}),
    (fun port ->
      if port < 0 then raise (Arg.Bad "") else
      master_mode := {!master_mode with Master.skl_port = port}) in
  let (set_program_path, set_color, set_link) =
    (fun path ->
      slave_mode := {!slave_mode with Slave.exe_path = path}),
    (fun col ->
      slave_mode := {!slave_mode with Slave.color = col}),
    (fun lnk ->
      slave_mode := {!slave_mode with Slave.link = lnk}) in
  let (set_local_addr) =
    (fun x ->
      master_mode :=
        {!master_mode with Master.local_addr = Unix.inet_addr_of_string x};
      slave_mode :=
        {!slave_mode with Slave.local_addr = Unix.inet_addr_of_string x}) in
  let arglist =
    [ "-master", Arg.Int mkmaster,
      "<int>  Start a master node with <int> slaves (MODE)"
    ; "-slave", Arg.String mkslave,
      "<addr>  Start a slave node connected to master node\n\
       \tat address <addr> (MODE)"
    ; "-slaveport", Arg.Int set_slave_port,
      "<port>  Set the Unix port number to be used to connect slaves (M)."
    ; "-sklport", Arg.Int set_skl_port,
      "<port>  Set the Unix port number used to communicate with\n\
       \tSklml programs (M)."
    ; "-sklpath", Arg.String set_program_path,
      "<path>  Set the path of the Sklml program run on the\n\
       \tcurrent slave node (S)"
    ; "-color", Arg.Int set_color,
      "<int>  Set the color of the current slave node (S)"
    ; "-dynlink", Arg.Bool set_link,
      "<bool>  Enable dynamic link with OCaml libraries listed in\n\
       \tenvironment variable SKLML_DLINK (S)"
    ; "-local", Arg.String set_local_addr,
      "<addr>  Set the local address in use (M) (S)"
    ; "-v",
      Arg.Unit
        (Upph_id_card.make_print_short_version Sklml_version.pph_id_card),
      " Print the short form of the version information and exit."
    ; "-version",
      Arg.Unit
        (Upph_id_card.make_print_long_version Sklml_version.pph_id_card),
      " Print the long form of the version information and exit."
    ] in
  try
    Arg.parse_argv argv arglist ignore (get_usage ());
    match !mode with
    | Some mode -> mode
    | None -> Spawnlib.fail "Mode is unspecified, see help using -help option"
  with
    | Arg.Bad _ -> Spawnlib.fail "Bad argument, see help using -help option"
    | Arg.Help msg -> Spawnlib.fail msg
    | Failure msg -> Spawnlib.fail msg
;;

let main () =
  let mode = parse_args Sys.argv in
  match mode with
  | Slave node -> Slave.start node
  | Master node -> Master.start node
;;

main ();;
