(***********************************************************************)
(*                                                                     *)
(*                                Sklml                                *)
(*                                                                     *)
(*                 The OCaml parallel skeleton system                  *)
(*                                                                     *)
(*               EPI Pomdapi, INRIA Paris - Rocquencourt               *)
(*                                                                     *)
(*  Copyright 2007-2014 INRIA.                                         *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved. This file is distributed only by permission.  *)
(*                                                                     *)
(*  Quentin Carbonneaux <Quentin.Carbonneaux@inria.fr>                 *)
(*  Francois Clement <Francois.Clement@inria.fr>                       *)
(*  Pierre Weis <Pierre.Weis@inria.fr>                                 *)
(*                                                                     *)
(***********************************************************************)

(* $Id: iotree.mli,v 1.17 2014/01/21 09:21:16 fclement Exp $ *)

(** {3 IO trees manipulations} *)

(** This modules manipulates IO trees, they are built in order to
  represent fork-join trees. Forks & joins create computation networks
  involving several nodes.  In order to send a stream in such networks
  functions are needed (for instance to preserve the order of data
  transmission), this is the purpose of this module. It has to be put
  in relation with the standard sequential code, especially the
  functions [send_in_tree] and [receive_in_tree]. *)

(** Here is a short example, the following skeleton : [map (seq
  identity, 2)] will be represented by the following computation tree
  ([cmp_tree] type) : [Par (Atom identity, Atom identity)]. With this
  description of the skeleton, a computation network will be spawned
  (here, two workers running the [identity] function). Once done, we
  have two addresses, [addr1] and [addr2], which will be put (by the
  [parfun] function) in the following IO tree : [IOABranch (IOALeaf
  addr1, IOAleaf addr2)]. Sending in this tree is made easy with this
  module juste create the tree [tree = iotree_of_ioatree (IOABranch (
  ... ))], this will have the side effect of connecting the two
  addresses with sockets, then use [send_in_tree tree] and
  [receive_in_tree tree]. *)

type ioatree =
   | IOABranch of ioatree * ioatree
   | IOASplit of ioatree * ioatree
   | IOASumSplit of ioatree * ioatree
   | IOALeafs of ioatree array
   | IOALeaf of Unix.sockaddr
   | IOALeafId
;;

type iotree =
   | IOBranch of iotree * iotree
   | IOSplit of iotree * iotree
   | IOSumSplit of iotree * iotree * (IO.ic * IO.oc)
   | IOLeafs of iotree array
   | IOLeaf of (IO.ic * IO.oc)
;;

val iotree_of_ioatree : ioatree -> iotree;;
val close_tree : iotree -> unit;;
val map_tree : iotree -> 'a Sklstream.t -> 'b Sklstream.t;;
val send_in_tree : iotree -> 'a Sklstream.t -> unit;;
val receive_in_tree : iotree -> 'b Sklstream.t;;
